\  oop.4th
\ -----------------------------------------------------------------------------
\  Classes, Objects and Methods for csp4th                           MM-070222
\ -----------------------------------------------------------------------------
\ 
\                     Copyright (C) 2008 Manfred Mahlow
\ 
\  This is Open Source Software released under the "Artistic License". See the
\  enclosed LICENSE file.
\ -----------------------------------------------------------------------------
 
  needs search-order
 
  only definitions  forth also  hidden also   decimal
 
  macro
  : .. (C -> .. ) (S -- )  \ compiling , interpreting
    (esm) off (osm) off ;
 
  : definitions (S -- )
    definitions (private) off  (esm) @  (ecm) ! ;
 
  ' ??  alias ??? (S -- )
 
  hidden definitions
 
  variable (c.root)      root  v.context @ (c.root) !
  variable (c.methods)   forth v.context @ (c.methods) !
  variable (c.macros)    forth v.context @ (c.macros) !
  variable (o.methods)   forth v.context @ (o.methods) !
  variable (o.macros)    forth v.context @ (o.macros) !
 
  variable c.link        0 c.link !
 
  \  The body of a class :
 
  \  +------------------------+
  \  |        wa(last)        |  addr = cid = context|current @
  \  +------------------------+
  \  |        wa(self)        |  *)
  \  +------------------------+
  \  | cid(previous context)  |
  \  +------------------------+
  \  | cid(container context) |
  \  +------------------------+
  \  | addr units per object  |
  \  +------------------------+
  \  | cid(inherited context) |
  \  +------------------------+
 
  \ *) bit 0 of wa is used as lock flag,
  \    bit 1 of wa is used as private flag
 
  : c.self ( cid -- a ) cell+ ;
 
  : c.container ( cid -- a ) [ 3 cells ] literal + ;
 
  : o.units  ( cid -- a )    [ 4 cells ] literal + ;
 
  : c.inherited ( cid -- a ) [ 5 cells ] literal + ; 
 
  : c.xt ( cid -- xt ) c.self @ [ %11 invert ] literal and >xt @ ;
 
  : c?defs ( -- )
    (ecm) @ 0if ." Error: (non-default) compile mode missing  " -1 throw then
  ;
 
  : c?objects ( cid -- )
      o.units @ 0if .cn ." Error: no object defined   " -1 throw then ;
 
  : o.seal ( cid -- )
      c.self dup @ %1 or swap ! ;
 
  : o?sealed ( cid -- )
      c.self @ %1 and 0; ." Error: Class is sealed   " -1 throw ;
 
  : c.prep-order ( cid1 .. cidu1 u1 cid -- cid1 ... cidu2 u2 )
    \ Prepend the class specific search order of the class context, cid 
    \ is pointing to, to the search order already present on the stack.
    0;  0 >r  \ set a mark on the returnstack
    begin >r 1+ r@ c.inherited @ dup 0until drop r>
    begin swap r> dup 0until drop ;
 
  : c.get-order ( -- c.root o.methods 2 )
                ( -- c.root o.macros 2 )
                ( -- c.root c.methods o.methods 3 )
                ( -- c.root c.macros  o.macros 3 )
    \ Build the state dependend search order part for a class 
    (c.root) @ prelude? 
    0if \ class was not executed as a prelude
       compiling?
       if (c.macros) @ (o.macros) else (c.methods) @ (o.methods) then @ 3
    else \ class was executed as a prelude
       compiling? if (o.macros) else (o.methods) then @ 2
    then ;
 
  : i.get-order ( -- c.root 1 )
                ( -- c.root c.methods 2 )
                ( -- c.root c.macros 2 )
    \ Build the state dependend search order part for an interface. An
    \ Interface is a class without an object definition ( units = 0 ).
    \ This is often called an abstract class.
    (c.root) @  (osm) @
    if    \ object search mode
      1
    else  \ no object search mode
      compiling? if (c.macros) else (c.methods) then @ 2
    then ;
 
  : c?get-order ( cid -- cid.1 ... cid.n n )
               o.units @ if c.get-order else i.get-order then (esm) on ;
 
  : c.set-order ( cid -- )
          (osm) @ 0if prelude? if (osm) on else (osm) off then then
          dup >r c?get-order r> c.prep-order set-order  ;
 
  forth definitions
 
  : class (S "name" -- )
    [compile] macro create
    here                               \ cid of the new class
    0 ,                                \ set empty wordlist
    current @ @ ,                      \ wa of the new class
    c.link @ ,  c.link !               \ add class to c.link
    (ecm) @ if current @ else 0 then , \ store container class
    0 ,                                \ no object defined (no address units)
    0 ,                                \ cid of the inherited class (none)
    does> ( cid -- ) \ build and set the class search order
          c.set-order ;
 
  ' class alias iface (S "name" -- )
 
  : .classes (S -- )
    c.link @ .clink
  ;
 
  : method (S -- )
    ['] .. prelude ;
 
  : m' (S "name" -- xt )
    ' method ;
 
  defer m: (S "name" -- )
        :: method [compile] :
         ; is m:
 
  macro
  : self (C -> current ) (S -- )  \ compiling , interpreting
    current @  (ecm) @ if  c.set-order  else  context !  then ;
 
  :: (S a u cid -- wa )     \ word found
     (S a u cid -- a u 0 )  \ word not found
     [ ' (c.find) defer@ ] compile, dup 0if ( a u 0 -- ) exit then ( wa -- )
     (osm) @ 0if exit then ( wa -- )
     dup >class @ %10 and if ( private ) >name count false then
  ; is (c.find)
 
  : private (S -- )
    c?defs %10 (private) ! ;
 
  : public (S -- )
    c?defs (private) off ;
 
\ : cb ( "name" xt numparams stacksize -- ) method cb ;

  : cb ( "name" xt numparams stacksize -- ) 
       (ecm) @ if private method cb public else cb then ;

  : stdcb ( "name" xt numparams stacksize --   ) method stdcb ;
 
  hidden ( interface method ) definitions
 
  class c.root    
  class c.methods 
  class c.macros  
  class o.methods 
  class o.macros  
 
  c.root .. e.context @ current !  (ecm) on   \ = c.root definitions
 
  current @ (c.root) !
 
  macro
  : self (C -> current ) (S -- )
    [compile] .. [compile] self ;
 
  macro
  ' (  alias ( (S -- )  \ Alias of the core word.

  macro
  ' [  alias [ (S -- )  \ Alias of the core word.
 
  ' ]  alias ] (S -- )  \ Alias of the core word.

  alias: \ | (S -- )
 
  macro
  ' ..  alias .. (S -- xt )  \ Alias of the word .. in the root vocabulary.
 
  ' .s  alias .s (S -- )     \ Alias of the stack printing word.
 
  ' words  alias words (S -- )  \ Alias of the core word.
 
  ' ??  alias ?? (S -- )
 
  : ??? (S -- )
    ??
    get-order
    begin
      cr " ----------------------------------------" 2dup type type
      previous words  e.context cell+ @ 
    0until 
    set-order ;
 
  macro
  : ; (S -- ) (C <iface> -> .. )
    compiling? 0if ." ; ?  compile mode missing  " -1 throw then
    last @ 0if ." ; ?  default search order missing  " -1 throw then
    [compile] ;
    e.context @ c.xt 1 or last @ >class
    dup >r @ %10 and or ( private ? )
        r> !  [compile] .. ;
 
   forth [DEFINED] h [IF] ' h alias h ( "name" -- ) [THEN]
 
  c.methods ..  e.context @ current !  (ecm) on  current @ (c.methods) !  
 
  m: definitions (S -- )  \ interpreting
     (esm) on  definitions  (esm) off ;
 
  : ' (S "name" -- xt ) (C <iface> -> .. )
    ' [compile] .. ;
 
  : m' (S "name" -- xt ) (C <iface> -> .. ) 
    self ' method ;
 
  : [DEFINED] (S "name" -- flag ) (C <iface> -> .. )
    [compile] [DEFINED]  [compile] .. ;
 
  ' reveal  alias reveal (S "name" -- ) \ Alias of the core word.
 
 
  private
  m: align (S -- )  \ interpreting
     c?defs  current @  dup o?sealed  o.units @  \ cr  dup ." old size : " .
     cell+ 1 -  1 cells /  cells                 \ cr  dup ." new size : " .
     current @ o.units ! ;
 
  private
  m: allot (S u -- )  \ interpreting
     c?defs current @  dup o?sealed  o.units +! ;
 
  private
  m: seal (S -- )  \ interpreting
    \ c?defs current @ dup o?sealed o.seal ;  MM-0802
     c?defs current @ dup o?sealed o.seal ;
 
  private
  m: units (S -- u )  \ interpreting
     e.context @ o.units @ ;
 
  private
  : inherit (S -- ) (C <class> -> .. )
    c?defs  e.context @  current @  dup o?sealed
    2dup = if drop .cn ." inherit ?  Can not inherit myself  " -1 throw then
    dup c.inherited @  over o.units @ or
    if over .cn ." inherit ?  Can not inherit any more  " -1 throw then
    over o.units @ over o.units !   \ inherit the object size
    2dup c.inherited !              \ inherit the class context
    drop o.seal [compile] .. ;      \ seal the inherited class
 
  private
  : super (C <class.1> -> <class.2> ) (S -- )
    e.context @ c.inherited @ 
    dup 0if ." super ?  No inherited class  " -1 throw then c.set-order ;
 
  macro
  : up (S -- ) (C <iface1> -> <iface2> )
    e.context @  c.container @
    dup 0if ." up ? no container context  " -1 throw then  c.set-order ;
 
  o.methods definitions   current @ (o.methods) !
 
  private
  : new (S "name" -- ) (C <class> --> .. )
    e.context @  dup c?objects  dup c.xt prelude  create
    dup o.units @ 0  do 0 c, loop  o.seal  [compile] .. ;
 
  private
  : this (S "name" -- ) (C <class> -> .. )
    c?defs  current @ o?sealed  e.context @ dup c?objects
    dup c.xt prelude  create  current @ o.units @ ,
    dup o.units @  current @ o.units +!  o.seal  [compile] ..
    does> ( oid a.offset -- oid' ) @ + ;
 
  public
  m: oid (S oid -- oid )
   ;
 
  c.macros definitions
 
  current @ (c.macros) !
 
  private macro
  c.methods ' super  alias super (C <class.1> -> <class.2> ) (S -- )
 
  private macro 
  : units (C <class> -> .. ) (S -- )  \ compiling
          (S -- u )                   \ executing
    [compile] ..  c.methods units  [compile] literal ;
 
  macro
  c.methods ' up  alias up (S -- ) (C <iface1> -> <iface2> )
 
  macro
  : ['] (S "name" -- ) (C <iface> -> .. )
    [compile] ['] [compile] .. ;
 
  o.macros definitions   current @ (o.macros) !
 
  public
  m: oid (S oid -- oid )
   ;
 
1 [IF]  \ Use object stack mechanism ? Yes !
 
  hidden ( object stack ) definitions
 
  variable (-os)  (-os) on   \ ON if the method 'this' is not supported in a
                             \ methods definition.
 
  variable (oid)  (oid) off  \ Top of the object stack (TOOS). The other 
                             \ values of the object stack are hold on the
                             \ return stack.
 
  : (this) ( -- oid )
  \ Return the TOOS on the data stack.
    (oid) @ ;
 
  : o.pop ( -- )
  \ Drop the TOOS
    r> (oid) ! ;
 
  : o.push ( oid -- )
  \ Push the oid from the TOS to TOOS.
    r> (oid) @ >r ['] o.pop >r >r (oid) ! ;

  c.root definitions
 
  private
  ' dup alias dup 
  \ Needed to hand over the oid of the calling object to methods of the
  \ called one.
  \ Usage:  this ( -- oid.1 ) dup <object> ( -- oid.1 oid.2 ) <method>
  \         ( -- oid.1 ) self dup <object> ( -- oid.1 oid.2 ) <method> 
 
  forth definitions

loc:

   : opush ( -- ) \ push TOOS to the return stack      MM-080324
      r> (oid) @ >r >r ;

   : opop ( -- ) \ pop TOOS from the return stack     MM-080324
      r> r> (oid) ! >r ;

   macro
   : catch ( xt -- n )
     \ The top of the object stack must be saved and restored when the 
     \ object stack is used in a method.
     compiling? 0if ." catch ? compile mode missing  "  -1 throw then
     (ecm) @ if ['] opush [compile] compile, then
     ['] catch [compile] compile,
     (ecm) @ if ['] opop [compile] compile, then ;

   reveal catch

loc;

  :: (S "name" --)
     [ ' m: defer@ ] compile,  here (-os) !
   ; is m: ( "name" --)
 
  : :: (S -- xt )
    [compile] :: here (-os) ! ;
 
  : : (S "name" -- )
    [compile] : here (-os) ! ;
 
  macro
  : this (C -> <class> ) (S -- )  \ compiling
         (S -- oid )              \ executing
    c?defs  compiling? 0if ." this ? Compile mode missing.  " -1 throw then  
    (-os) @ if ." this ?  Object stack underflow.  " -1 throw then
    current @ c.set-order  ['] (this) [compile] compile, ;
 
  macro
  : (O (S -- ) \ compiling
       (S oid -- ) \ executing , moves TOS to TOOS
    c?defs  compiling? 0if ." Error: Compile mode missing  " -1 throw then
    (-os) @ here = 0if ." (O ? invalid use of the object stack  " -1 throw then
    ['] o.push  [compile] compile,  (-os) off  [compile] ( ;
 
[THEN]
 
 hidden definitions
 
  hide (c.root)  hide (c.methods)  hide (c.macros)  hide (o.methods)
  hide (o.macros)
 
  forth definitions
 
  verbose @ not [IF] hide (private)  hide (prelude) [THEN]
 
  vocabulary oop (S -- )
 
  only  forth also  oop definitions also
 
\ -----------------------------------------------------------------------------
\  Last revision : MM-071227
\                  c.create removed
\                  MM-071226
\                  'c.xt execute' replaced by 'c.set-order'
\                  MM-071223
\                  class modified to make iface.src obsolet
\                  A class without data ( units = 0 ) is an interface or
\                  abstract class. It can inherit from another interface or
\                  class as long as it is not sealed.
\                  MM-080108
\                  h added to c.root
\                  MM-080208
\                  MM-080303  cb changed to create private callback handlers in
\                             class and iface contexts
\                  MM-080324  catch redefined to save and restore the TOOS
\                  MM-080409  stack diagrams
\                  MM-080420  \ added to c.root
\                  MM-080530  Reva repeat while -> begin 0until
